from .pin import Pin

class Chip:
    OUTPUT_PINS = []
    INPUT_PINS = []
    STARTING_HIGH = [] # pins that start high
    PIN_ORDER = None

    def __init__(self):
        for code in self.OUTPUT_PINS:
            pin = Pin(code, chip=self, output=True, high=(code in self.STARTING_HIGH))
            setattr(self, 'pin_{}'.format(code), pin)
        for code in self.INPUT_PINS:
            pin = Pin(code, chip=self, high=(code in self.STARTING_HIGH))
            setattr(self, 'pin_{}'.format(code), pin)
        self.vcc = Pin('VCC', chip=self, high=True)
        self.update()

    def __str__(self):
        inputs = ' '.join(str(self.getpin(code)) for code in self.INPUT_PINS)
        outputs = ' '.join(str(self.getpin(code)) for code in self.OUTPUT_PINS)
        return '{} I: {} O: {}'.format(self.__class__.__name__, inputs, outputs)

    def asciiart(self):
        if self.PIN_ORDER:
            pin_order = self.PIN_ORDER
        else:
            pin_order = self.INPUT_PINS + self.OUTPUT_PINS
        pins = list(self.getpins(pin_order))
        if len(pins) % 2:
            pins.append(None)
        left_pins = pins[:len(pins) // 2]
        right_pins = reversed(pins[len(pins) // 2:])
        lines = []
        lines.append('     _______     ')
        for index, (left, right) in enumerate(zip(left_pins, right_pins)):
            larrow = '<' if left.output else '>'
            lpol = '+' if left.ishigh() else '-'
            sleft = left.code.rjust(4) + larrow + '|' + lpol
            if right:
                rarrow = '>' if right.output else '<'
                rpol = '+' if right.ishigh() else '-'
                sright = rpol + '|' + rarrow + right.code.ljust(4)
            else:
                sright = ' |     '

            if index == len(left_pins) - 1:
                spacer = '_'
            else:
                spacer = ' '
            middle = 'U' if index == 0 else spacer
            line = sleft + spacer + middle + spacer + sright
            lines.append(line)
        return '\n'.join(lines)

    def ispowered(self):
        return self.vcc.ishigh()

    def getpin(self, code):
        return getattr(self, 'pin_{}'.format(code))

    def getpins(self, codes):
        return (self.getpin(code) for code in codes)

    # Set multiple pins on the same chip and only update chips one all pins are updated.
    def setpins(self, low, high):
        updateself = False
        updatelist = set()
        for codes, val in [(low, False), (high, True)]:
            for pin in self.getpins(codes):
                if pin.high != val:
                    pin.high = val
                    if pin.output:
                       updatelist |= pin.propagate_to()
                    else:
                        updateself = True
        if updateself:
            self.update()
        for chip in updatelist:
            chip.update()

    def tick(self, us):
        pass

    def update(self):
        pass

    # Same as with setpins, but for wire_to()
    # Has to be called from the chip having the *input* pins
    def wirepins(self, chip, inputs, outputs):
        for icode, ocode in zip(inputs, outputs):
           ipin = self.getpin(icode)
           opin = chip.getpin(ocode)
           assert opin.output
           assert not ipin.output
           ipin.wires.add(opin)
           opin.wires.add(ipin)
        self.update()

class ActivableChip(Chip):
    ENABLE_PINS = [] # ~ means that low == enabled

    def is_enabled(self):
        for code in self.ENABLE_PINS:
            pin = self.getpin(code.replace('~', ''))
            enabled = pin.ishigh()
            if code.startswith('~'):
                enabled = not enabled
            if not enabled:
                return False
        return True




# first pin is the least significant bit
def set_binary_value(value, pins):
    for i, pin in enumerate(pins):
        pin.set(bool((value >> i) & 0x1))

def get_binary_value(pins):
    value = 0
    for index, pin in enumerate(pins):
        if pin.ishigh():
            value |= 1 << index
    return value




from .chip import Chip

class Inverter(Chip):
    def update(self):
        for in_, out in zip(self.INPUT_PINS, self.OUTPUT_PINS):
            pin_in = self.getpin(in_)
            pin_out = self.getpin(out)
            pin_out.set(not pin_in.ishigh())


class SN74HC14(Inverter):
    INPUT_PINS = ['1A', '2A', '3A', '4A', '5A', '6A']
    OUTPUT_PINS = ['1Y', '2Y', '3Y', '4Y', '5Y', '6Y']




from .chip import Chip

class LED:
    def __init__(self, vcc, gnd, fade_delay_us=10000):
        self.vcc = vcc
        self.gnd = gnd
        self.fade_delay_us = fade_delay_us
        self.fade_counter_us = fade_delay_us + 1

    def tick(self, us):
        self.fade_counter_us += us
        self.ishigh()

    def ishigh(self):
        if self.vcc.ishigh() and not self.gnd.ishigh():
            self.fade_counter_us = 0
            return True
        else:
            return self.fade_counter_us <= self.fade_delay_us


def combine_repr(*segs):
    """Combine and returns __str__ repr of multiple Segment7
    """
    outputs = [str(seg) for seg in segs]
    line1 = ' '.join(s[:3] for s in outputs)
    line2 = ' '.join(s[4:7] for s in outputs)
    line3 = ' '.join(s[8:] for s in outputs)
    return '\n'.join([line1, line2, line3])

class Segment7(Chip):
    INPUT_PINS = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'DP']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.leds = {pin.code: LED(self.vcc, pin) for pin in self.getpins(self.INPUT_PINS)}

    def __str__(self):
        SEGMENTS = """
 _.
|_|
|_|""".strip('\n')
        # Remember, each line is **4** chars long if we count the \n!
        SEGPOS = [
            '', 'A', 'DP', '',
            'F', 'G', 'B', '',
            'E', 'D', 'C'
        ]
        return ''.join(c if seg and self.leds[seg].ishigh() else ' ' for c, seg in zip(SEGMENTS, SEGPOS))

    def tick(self, us):
        for led in self.leds.values():
            led.tick(us)




import curses
import time

class UIElement:
    def __init__(self, label, outputfunc):
        self.label = label
        self.outputfunc = outputfunc

class UIAction:
    def __init__(self, key, label, func):
        self.key = key
        self.label = label
        self.func = func

class UIScreen:
    def __init__(self, refresh_rate_us=0):
        self.stdscr = curses.initscr()
        self.refresh_rate_us = refresh_rate_us
        self.refresh_counter = 0
        curses.noecho()
        curses.cbreak()
        self.stdscr.nodelay(True)
        self.elements = []
        self.actions = []
        self.key2actions = {}
        self.last_ch = -1
        self.elements_win = curses.newwin(1, 42, 0, 0)
        self.action_win = curses.newwin(1, 42, 1, 1)

    def _element_lines(self):
        lines = []
        for elem in self.elements:
            lines += [elem.label]
            lines += elem.outputfunc().splitlines()
        return lines

    def _read_key(self):
        ch = self.stdscr.getch()
        if ch > 0 and ch != self.last_ch:
            key = chr(ch)
            if key in self.key2actions:
                self.key2actions[key].func()
        self.last_ch = ch

    def _refresh_actions(self):
        if not self.actions:
            return
        win = self.action_win
        win.erase()
        _, w = win.getmaxyx()
        win.addnstr(0, 0, "Menu:", w)
        y = 1
        for action in self.actions:
            win.addnstr(y, 0, "{} - {}".format(action.key, action.label), w)
            y+= 1
        win.refresh()

    def _refresh_elements(self):
        self.last_refresh = time.time()
        win = self.elements_win
        win.erase()
        _, w = win.getmaxyx()
        for y, line in enumerate(self._element_lines()):
            win.addnstr(y, 0, line, w)
        win.refresh()

    def _resize_windows(self):
        h, w = self._win_elements_size()
        self.elements_win.resize(h, w)
        self.action_win.mvwin(h + 2, 0)
        self.action_win.resize(*self._win_actions_size())
        self.last_refresh = 0
        self.refresh()

    def _win_elements_size(self):
        lines = self._element_lines()
        h = len(lines)
        # Let's give ourselves a little buffer of 4 chars...
        w = max((len(line) for line in lines), default=1) + 4
        return (h, w)

    def _win_actions_size(self):
        h = len(self.actions) + 1
        # 5 chars are for key prefixes
        w = max((len(a.label) for a in self.actions), default=1) + 5
        return (h, w)

    def tick(self, us):
        self.refresh_counter += us

    def stop(self):
        curses.nocbreak()
        curses.echo()
        curses.endwin()

    def add_element(self, label, outputfunc):
        self.elements.append(UIElement(label, outputfunc))
        self._resize_windows()

    def add_action(self, key, label, func):
        self.actions.append(UIAction(key, label, func))
        self.key2actions = {a.key: a for a in self.actions}
        self._resize_windows()
        self._refresh_actions()

    def refresh(self):
        if self.refresh_counter >= self.refresh_rate_us:
            self.refresh_counter = 0
            self._refresh_elements()
            self.stdscr.refresh()

        self._read_key()




from .chip import ActivableChip
from .util import set_binary_value

class BinaryCounter(ActivableChip):
    CLOCK_PIN = ''

    def __init__(self, *args, **kwargs):
        self.value = 0
        self.prev_clock_high = False
        super().__init__(*args, **kwargs)

    def maxvalue(self):
        return (1 << len(self.OUTPUT_PINS)) - 1

    def update(self):
        if not self.is_enabled():
            return

        clock = self.getpin(self.CLOCK_PIN)
        if clock.ishigh() and not self.prev_clock_high:
            self.value += 1
            if self.value > self.maxvalue():
                self.value = 0

            set_binary_value(self.value, self.getpins(self.OUTPUT_PINS))

        self.prev_clock_high = clock.ishigh()


class SN74F161AN(BinaryCounter):
    CLOCK_PIN = 'CLK'
    ENABLE_PINS = ['ENT', 'ENP']
    STARTING_HIGH = ENABLE_PINS
    INPUT_PINS = [CLOCK_PIN] + ENABLE_PINS
    OUTPUT_PINS = ['QA', 'QB', 'QC', 'QD']




class Pin:
    def __init__(self, code, high=False, chip=None, output=False):
        self.code = code
        self.high = high
        self.chip = chip
        self.output = output
        self.wires = set()

    def __str__(self):
        return '{}/{}'.format(self.code, 'H' if self.ishigh() else 'L')

    def ishigh(self):
        if not self.output and self.wires:
            wired_outputs = (p for p in self.wires if p.output)
            return any(p.ishigh() for p in wired_outputs)
        else:
            return self.high

    def propagate_to(self):
        if self.output:
            return {
                p.chip for p in self.wires
                if not p.output and p.chip is not self and p.chip is not None
            }
        else:
            return set()

    def set(self, val):
        if val == self.high:
            return

        self.high = val
        if not self.output and self.chip:
            self.chip.update()

        if self.output:
            wired_chips = self.propagate_to()
            for chip in wired_chips:
                chip.update()

    def sethigh(self):
        self.set(True)

    def setlow(self):
        self.set(False)

    def toggle(self):
        self.set(not self.ishigh())

    def wire_to(self, output_pin):
        assert not self.output
        assert output_pin.output
        self.wires.add(output_pin)
        output_pin.wires.add(self)
        if self.chip:
            self.chip.update()

def pinrange(prefix, start, end):
    # pinrange('Y', 0, 3) -> ['Y0', 'Y1', 'Y2', 'Y3']
    return [prefix + str(i) for i in range(start, end+1)]



from .chip import ActivableChip

class Decoder(ActivableChip):
    SERIAL_PINS = [] # LSB pin goes first
    RESULT_PINS = [] # LSB pin goes first

    def update(self):
        selection = 0 if self.is_enabled() else 0xff
        for index, code in enumerate(self.SERIAL_PINS):
            pin = self.getpin(code)
            selection |= int(pin.ishigh()) << index
        for index, code in enumerate(self.RESULT_PINS):
            pin = self.getpin(code)
            pin.set(index != selection)


class SN74HC138(Decoder):
    OUTPUT_PINS = ['Y0', 'Y1', 'Y2', 'Y3', 'Y4', 'Y5', 'Y6', 'Y7']
    INPUT_PINS = ['A', 'B', 'C', 'G2A', 'G2B', 'G1']
    SERIAL_PINS = ['A', 'B', 'C']
    RESULT_PINS = OUTPUT_PINS
    ENABLE_PINS = ['G1', '~G2A', '~G2B']
    STARTING_HIGH = ['G1'] + RESULT_PINS




from .chip import ActivableChip
from .util import set_binary_value

class ShiftRegister(ActivableChip):
    SERIAL_PINS = []
    CLOCK_PIN = ''
    BUFFER_PIN = '' # Pin that makes the SR buffer go to output pin. Leave blank if there's none
    RESULT_PINS = [] # Data is pushed from first pin to last

    def __init__(self, *args, **kwargs):
        self.prev_clock_high = False
        self.prev_buffer_high = False
        self.buffer = 0
        super().__init__(*args, **kwargs)

    def update(self):
        if not self.is_enabled():
            self.setpins(self.RESULT_PINS, False)
            return

        newbuffer = self.buffer

        clock = self.getpin(self.CLOCK_PIN)
        if clock.ishigh() and not self.prev_clock_high:
            newbuffer = newbuffer << 1
            if all(self.getpin(code).ishigh() for code in self.SERIAL_PINS):
                newbuffer |= 0x1
        self.prev_clock_high = clock.ishigh()

        should_refresh_outputs = False
        if self.BUFFER_PIN:
            bufpin = self.getpin(self.BUFFER_PIN)
            should_refresh_outputs = bufpin.ishigh() and not self.prev_buffer_high
            self.prev_buffer_high = bufpin.ishigh()
            # we don't set self.buffer because, per design, buffered SRs suffer a delay
            # when the buffer pin is activated at the same time as the clock pin.
        else:
            # if there's not buffering, there's no delay
            self.buffer = newbuffer
            should_refresh_outputs = True
        if should_refresh_outputs:
            set_binary_value(self.buffer, self.getpins(self.RESULT_PINS))

        self.buffer = newbuffer


class CD74AC164(ShiftRegister):
    OUTPUT_PINS = ['Q0', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6', 'Q7']
    INPUT_PINS = ['DS1', 'DS2', 'CP', 'MR']
    SERIAL_PINS = ['DS1', 'DS2']
    RESULT_PINS = OUTPUT_PINS
    ENABLE_PINS = ['MR']
    CLOCK_PIN = 'CP'
    STARTING_HIGH = ['MR', 'DS2']


class SN74HC595(ShiftRegister):
    OUTPUT_PINS = ['QA', 'QB', 'QC', 'QD', 'QE', 'QF', 'QG', 'QH']
    INPUT_PINS = ['OE', 'RCLK', 'SER', 'SRCLK', 'SRCLR']
    SERIAL_PINS = ['SER']
    RESULT_PINS = OUTPUT_PINS
    ENABLE_PINS = ['~OE', 'SRCLR']
    CLOCK_PIN = 'SRCLK'
    BUFFER_PIN = 'RCLK'
    STARTING_HIGH = ['SRCLR']




from setuptools import setup

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Software Development :: Embedded Systems',
    'Topic :: System :: Emulators',
]

LONG_DESCRIPTION = open('README.md', 'rt').read()

setup(
    name='icemu',
    version='0.2.0',
    author='Virgil Dupras',
    author_email='hsoft@hardcoded.net',
    packages=['icemu'],
    scripts=[],
    url='https://github.com/hsoft/icemu',
    license='LGPLv3',
    description='Emulate Integrated Circuits',
    long_description=LONG_DESCRIPTION,
    classifiers=CLASSIFIERS,
)

